package database

import (
	"context"
	"errors"

	"git.anthrove.art/Anthrove/otter-space-sdk/v2/internal/utils"
	otterError "git.anthrove.art/Anthrove/otter-space-sdk/v2/pkg/error"
	"git.anthrove.art/Anthrove/otter-space-sdk/v2/pkg/models"
	log "github.com/sirupsen/logrus"
	"go.opentelemetry.io/otel/attribute"
	"gorm.io/gorm"
)

func CreateTag(ctx context.Context, tagName models.TagName, tagType models.TagType) (models.Tag, error) {
	ctx, span := tracer.Start(ctx, "CreateTag")
	defer span.End()

	span.SetAttributes(
		attribute.String("tag_name", string(tagName)),
		attribute.String("tag_type", string(tagType)),
	)

	span.AddEvent("Starting tag creation")

	if client == nil {
		return models.Tag{}, utils.HandleError(ctx, span, logger, nil, &otterError.Database{Reason: otterError.DatabaseIsNotConnected})
	}

	tag := models.Tag{
		Name: tagName,
		Type: tagType,
	}

	logger.WithContext(ctx).WithFields(log.Fields{
		"tag_name": tagName,
		"tag_type": tagType,
	}).Debug("attempting to create tag")

	result := client.WithContext(ctx).Create(&tag)
	if result.Error != nil {
		if errors.Is(result.Error, gorm.ErrDuplicatedKey) {

			loggerFields := log.Fields{
				"tag_name": tagName,
				"tag_type": tagType,
			}
			return models.Tag{}, utils.HandleError(ctx, span, logger, loggerFields, &otterError.Database{Reason: otterError.DuplicateKey})
		}

		loggerFields := log.Fields{
			"tag_name": tagName,
			"tag_type": tagType,
		}
		return models.Tag{}, utils.HandleError(ctx, span, logger, loggerFields, result.Error)
	}

	logger.WithContext(ctx).WithFields(log.Fields{
		"tag_name": tagName,
		"tag_type": tagType,
	}).Debug("tag created")
	span.AddEvent("Tag created successfully")
	return tag, nil
}

func CreateTagInBatch(ctx context.Context, tags []models.Tag, batchSize int) error {
	ctx, span := tracer.Start(ctx, "CreateTagInBatch")
	defer span.End()

	span.SetAttributes(
		attribute.Int64("batch_size", int64(batchSize)),
		attribute.Int64("tag_count", int64(len(tags))),
	)

	span.AddEvent("Starting batch tag creation")

	if client == nil {
		return utils.HandleError(ctx, span, logger, nil, &otterError.Database{Reason: otterError.DatabaseIsNotConnected})
	}

	if tags == nil || len(tags) == 0 {
		return utils.HandleError(ctx, span, logger, nil, &otterError.EntityValidationFailed{Reason: otterError.TagListIsEmpty})
	}

	if batchSize == 0 {
		return utils.HandleError(ctx, span, logger, nil, &otterError.EntityValidationFailed{Reason: otterError.BatchSizeIsEmpty})
	}

	logger.WithContext(ctx).WithFields(log.Fields{
		"tag_length": len(tags),
	}).Debug("attempting to create tags")

	result := client.WithContext(ctx).CreateInBatches(&tags, batchSize)
	if result.Error != nil {
		if errors.Is(result.Error, gorm.ErrDuplicatedKey) {
			loggerFields := log.Fields{
				"tag_length": len(tags),
			}
			return utils.HandleError(ctx, span, logger, loggerFields, &otterError.Database{Reason: otterError.DuplicateKey})
		}

		loggerFields := log.Fields{
			"tag_length": len(tags),
		}
		return utils.HandleError(ctx, span, logger, loggerFields, result.Error)
	}

	logger.WithContext(ctx).WithFields(log.Fields{
		"tag_length": len(tags),
	}).Debug("tags created")

	span.AddEvent("Batch tags created successfully")
	return nil
}

func DeleteTag(ctx context.Context, tagName models.TagName) error {
	ctx, span := tracer.Start(ctx, "DeleteTag")
	defer span.End()

	span.SetAttributes(
		attribute.String("tag_name", string(tagName)),
	)

	span.AddEvent("Starting tag deletion")

	var tag models.Tag

	if client == nil {
		return utils.HandleError(ctx, span, logger, nil, &otterError.Database{Reason: otterError.DatabaseIsNotConnected})
	}

	if len(tagName) == 0 {
		return utils.HandleError(ctx, span, logger, nil, &otterError.Database{Reason: otterError.TagNameIsEmpty})
	}

	logger.WithContext(ctx).WithFields(log.Fields{
		"tag_name": tagName,
	}).Debug("attempting to delete tag")

	result := client.WithContext(ctx).Delete(&tag, tagName)
	if result.Error != nil {
		if errors.Is(result.Error, gorm.ErrRecordNotFound) {

			loggerFields := log.Fields{
				"tag_name": tagName,
			}
			return utils.HandleError(ctx, span, logger, loggerFields, &otterError.Database{Reason: otterError.NoDataFound})
		}

		loggerFields := log.Fields{
			"tag_name": tagName,
		}
		return utils.HandleError(ctx, span, logger, loggerFields, result.Error)
	}

	logger.WithContext(ctx).WithFields(log.Fields{
		"tag_name": tagName,
	}).Debug("tag deleted")

	span.AddEvent("Tag deleted successfully")
	return nil
}
